@@ -385,10 +385,14 @@ next_while:;
      to create the timer once per thread.  Since WFMO checks the handles
      in order, we append the timer as last object, otherwise it's preferred
      over actual events on the descriptors. */
-  HANDLE &wait_timer = _my_tls.locals.cw_timer;
+  HANDLE local_timer = NULL;
+  HANDLE &wait_timer =
+    _my_tls.locals.cw_timer_inuse ? local_timer : _my_tls.locals.cw_timer;
   if (us > 0LL)
     {
       NTSTATUS status;
+      if (!_my_tls.locals.cw_timer_inuse)
+	_my_tls.locals.cw_timer_inuse = true;
       if (!wait_timer)
 	{
 	  status = NtCreateTimer (&wait_timer, TIMER_ALL_ACCESS, NULL,
@@ -431,6 +435,10 @@ next_while:;
     {
       BOOLEAN current_state;
       NtCancelTimer (wait_timer, &current_state);
+      if (local_timer)
+	NtClose (local_timer);
+      else
+	_my_tls.locals.cw_timer_inuse = false;
     }
 
   wait_states res;
@@ -584,14 +592,15 @@ no_verify (select_record *, fd_set *, fd_set *, fd_set *)
   return 0;
 }
 
-static int
-pipe_data_available (int fd, fhandler_base *fh, HANDLE h, bool writing)
+ssize_t
+pipe_data_available (int fd, fhandler_base *fh, HANDLE h, int flags)
 {
   if (fh->get_device () == FH_PIPER)
     {
       DWORD nbytes_in_pipe;
-      if (!writing && PeekNamedPipe (h, NULL, 0, NULL, &nbytes_in_pipe, NULL))
-	return nbytes_in_pipe > 0;
+      if (!(flags & PDA_WRITE)
+	  && PeekNamedPipe (h, NULL, 0, NULL, &nbytes_in_pipe, NULL))
+	return nbytes_in_pipe;
       return -1;
     }
 
@@ -609,9 +618,17 @@ pipe_data_available (int fd, fhandler_base *fh, HANDLE h, bool writing)
 	 access on the write end.  */
       select_printf ("fd %d, %s, NtQueryInformationFile failed, status %y",
 		     fd, fh->get_name (), status);
-      return writing ? 1 : -1;
+      switch (flags)
+	{
+	case PDA_WRITE:
+	  return 1;
+	case PDA_SELECT | PDA_WRITE:
+	  return PIPE_BUF;
+	default:
+	  return -1;
+	}
     }
-  if (writing)
+  if (flags & PDA_WRITE)
     {
       /* If there is anything available in the pipe buffer then signal
         that.  This means that a pipe could still block since you could
@@ -638,20 +655,26 @@ pipe_data_available (int fd, fhandler_base *fh, HANDLE h, bool writing)
       /* Note: Do not use NtQueryInformationFile() for query_hdl because
 	 NtQueryInformationFile() seems to interfere with reading pipes
 	 in non-cygwin apps. Instead, use PeekNamedPipe() here. */
+      /* Note 2: we return the number of available bytes.  Select for writing
+         returns writable *only* if at least PIPE_BUF bytes are left in the
+	 buffer.  If we can't fetch the real number of available bytes, the
+	 number of bytes returned depends on the caller.  For select we return
+	 PIPE_BUF to fake writability, for writing we return 1 to allow
+	 handling this fact. */
       if (fh->get_device () == FH_PIPEW && fpli.WriteQuotaAvailable == 0)
 	{
 	  HANDLE query_hdl = ((fhandler_pipe *) fh)->get_query_handle ();
 	  if (!query_hdl)
 	    query_hdl = ((fhandler_pipe *) fh)->temporary_query_hdl ();
-	  if (!query_hdl)
-	    return 1; /* We cannot know actual write pipe space. */
+	  if (!query_hdl) /* We cannot know actual write pipe space. */
+	    return (flags & PDA_SELECT) ? PIPE_BUF : 1;
 	  DWORD nbytes_in_pipe;
 	  BOOL res =
 	    PeekNamedPipe (query_hdl, NULL, 0, NULL, &nbytes_in_pipe, NULL);
 	  if (!((fhandler_pipe *) fh)->get_query_handle ())
 	    CloseHandle (query_hdl); /* Close temporary query_hdl */
-	  if (!res)
-	    return 1;
+	  if (!res) /* We cannot know actual write pipe space. */
+	    return (flags & PDA_SELECT) ? PIPE_BUF : 1;
 	  fpli.WriteQuotaAvailable = fpli.InboundQuota - nbytes_in_pipe;
 	}
       if (fpli.WriteQuotaAvailable > 0)
@@ -659,7 +682,7 @@ pipe_data_available (int fd, fhandler_base *fh, HANDLE h, bool writing)
 	  paranoid_printf ("fd %d, %s, write: size %u, avail %u", fd,
 			   fh->get_name (), fpli.InboundQuota,
 			   fpli.WriteQuotaAvailable);
-	  return 1;
+	  return fpli.WriteQuotaAvailable;
 	}
       /* TODO: Buffer really full or non-Cygwin reader? */
     }
@@ -667,7 +690,7 @@ pipe_data_available (int fd, fhandler_base *fh, HANDLE h, bool writing)
     {
       paranoid_printf ("fd %d, %s, read avail %u", fd, fh->get_name (),
 		       fpli.ReadDataAvailable);
-      return 1;
+      return fpli.ReadDataAvailable;
     }
   if (fpli.NamedPipeState & FILE_PIPE_CLOSING_STATE)
     return -1;
@@ -715,10 +738,10 @@ peek_pipe (select_record *s, bool from_select)
 	  gotone = s->read_ready = true;
 	  goto out;
 	}
-      int n = pipe_data_available (s->fd, fh, h, false);
+      ssize_t n = pipe_data_available (s->fd, fh, h, PDA_SELECT);
       /* On PTY masters, check if input from the echo pipe is available. */
       if (n == 0 && fh->get_echo_handle ())
-	n = pipe_data_available (s->fd, fh, fh->get_echo_handle (), false);
+	n = pipe_data_available (s->fd, fh, fh->get_echo_handle (), PDA_SELECT);
 
       if (n < 0)
 	{
@@ -759,9 +782,9 @@ out:
 	    gotone += s->except_ready = true;
 	  return gotone;
 	}
-      int n = pipe_data_available (s->fd, fh, h, true);
+      ssize_t n = pipe_data_available (s->fd, fh, h, PDA_SELECT | PDA_WRITE);
       select_printf ("write: %s, n %d", fh->get_name (), n);
-      gotone += s->write_ready = n;
+      gotone += s->write_ready = (n >= PIPE_BUF);
       if (n < 0 && s->except_selected)
 	gotone += s->except_ready = true;
     }
@@ -770,7 +793,7 @@ out:
 
 static int start_thread_pipe (select_record *me, select_stuff *stuff);
 
-static DWORD WINAPI
+static DWORD
 thread_pipe (void *arg)
 {
   select_pipe_info *pi = (select_pipe_info *) arg;
@@ -950,7 +973,17 @@ peek_fifo (select_record *s, bool from_select)
 	    }
 	}
       fh->fifo_client_unlock ();
-      if (!nconnected && fh->hit_eof ())
+      /* According to POSIX and the Linux man page, we're supposed to
+	 report read ready if the FIFO is at EOF, i.e., if the pipe is
+	 empty and there are no writers.  But there seems to be an
+	 undocumented exception, observed on Linux and other platforms
+	 (https://cygwin.com/pipermail/cygwin/2022-September/252223.html):
+	 If no writer has ever been opened, then we do not report read
+	 ready.  This can happen if a reader is opened with O_NONBLOCK
+	 before any writers have opened.  To be consistent with other
+	 platforms, we use a special EOF test that returns false if
+	 there's never been a writer opened. */
+      if (!nconnected && fh->select_hit_eof ())
 	{
 	  select_printf ("read: %s, saw EOF", fh->get_name ());
 	  gotone += s->read_ready = true;
@@ -962,9 +995,10 @@ peek_fifo (select_record *s, bool from_select)
 out:
   if (s->write_selected)
     {
-      int n = pipe_data_available (s->fd, fh, fh->get_handle (), true);
+      ssize_t n = pipe_data_available (s->fd, fh, fh->get_handle (),
+				       PDA_SELECT | PDA_WRITE);
       select_printf ("write: %s, n %d", fh->get_name (), n);
-      gotone += s->write_ready = n;
+      gotone += s->write_ready = (n >= PIPE_BUF);
       if (n < 0 && s->except_selected)
 	gotone += s->except_ready = true;
     }
@@ -973,7 +1007,7 @@ out:
 
 static int start_thread_fifo (select_record *me, select_stuff *stuff);
 
-static DWORD WINAPI
+static DWORD
 thread_fifo (void *arg)
 {
   select_fifo_info *pi = (select_fifo_info *) arg;
@@ -1131,7 +1165,9 @@ peek_console (select_record *me, bool)
       else
 	{
 	  acquire_attach_mutex (mutex_timeout);
+	  DWORD resume_pid = fh->attach_console (fh->get_owner ());
 	  BOOL r = PeekConsoleInputW (h, &irec, 1, &events_read);
+	  fh->detach_console (resume_pid, fh->get_owner ());
 	  release_attach_mutex ();
 	  if (!r || !events_read)
 	    break;
@@ -1161,7 +1197,7 @@ verify_console (select_record *me, fd_set *rfds, fd_set *wfds,
 
 static int console_startup (select_record *me, select_stuff *stuff);
 
-static DWORD WINAPI
+static DWORD
 thread_console (void *arg)
 {
   select_console_info *ci = (select_console_info *) arg;
@@ -1351,8 +1387,6 @@ peek_pty_slave (select_record *s, bool from_select)
   fhandler_base *fh = (fhandler_base *) s->fh;
   fhandler_pty_slave *ptys = (fhandler_pty_slave *) fh;
 
-  ptys->reset_switch_to_pcon ();
-
   if (s->read_selected)
     {
       if (s->read_ready)
@@ -1388,9 +1422,9 @@ out:
   HANDLE h = ptys->get_output_handle ();
   if (s->write_selected)
     {
-      int n = pipe_data_available (s->fd, fh, h, true);
+      ssize_t n = pipe_data_available (s->fd, fh, h, PDA_SELECT | PDA_WRITE);
       select_printf ("write: %s, n %d", fh->get_name (), n);
-      gotone += s->write_ready = n;
+      gotone += s->write_ready = (n >= PIPE_BUF);
       if (n < 0 && s->except_selected)
 	gotone += s->except_ready = true;
     }
@@ -1399,7 +1433,7 @@ out:
 
 static int pty_slave_startup (select_record *me, select_stuff *stuff);
 
-static DWORD WINAPI
+static DWORD
 thread_pty_slave (void *arg)
 {
   select_pipe_info *pi = (select_pipe_info *) arg;
@@ -1437,7 +1471,7 @@ pty_slave_startup (select_record *me, select_stuff *stuff)
   fhandler_base *fh = (fhandler_base *) me->fh;
   fhandler_pty_slave *ptys = (fhandler_pty_slave *) fh;
   if (me->read_selected)
-    ptys->mask_switch_to_pcon_in (true, true);
+    ptys->mask_switch_to_nat_pipe (true, true);
 
   select_pipe_info *pi = stuff->device_specific_ptys;
   if (pi->start)
@@ -1464,7 +1498,7 @@ pty_slave_cleanup (select_record *me, select_stuff *stuff)
   if (!pi)
     return;
   if (me->read_selected && pi->start)
-    ptys->mask_switch_to_pcon_in (false, false);
+    ptys->mask_switch_to_nat_pipe (false, false);
   if (pi->thread)
     {
       pi->stop_thread = true;
@@ -1775,7 +1809,7 @@ peek_socket (select_record *me, bool)
 
 static int start_thread_socket (select_record *, select_stuff *);
 
-static DWORD WINAPI
+static DWORD
 thread_socket (void *arg)
 {
   select_socket_info *si = (select_socket_info *) arg;
@@ -2245,3 +2279,145 @@ fhandler_timerfd::select_except (select_stuff *stuff)
   s->except_ready = false;
   return s;
 }
+
+static int
+peek_dsp (select_record *s, bool from_select)
+{
+  int gotone = 0;
+  fhandler_dev_dsp *fh = (fhandler_dev_dsp *)(fhandler_base *) s->fh;
+
+  if (s->read_selected)
+      if (s->read_ready || fh->read_ready ())
+	gotone += s->read_ready = true;
+  if (s->write_selected)
+      if (s->write_ready || fh->write_ready ())
+	gotone += s->write_ready = true;
+  if (s->except_selected)
+      if (s->except_ready || fh->is_closed ())
+	gotone += s->except_ready = true;
+  return gotone;
+}
+
+static int start_thread_dsp (select_record *me, select_stuff *stuff);
+
+static DWORD
+thread_dsp (void *arg)
+{
+  select_dsp_info *di = (select_dsp_info *) arg;
+  DWORD sleep_time = 0;
+  bool looping = true;
+
+  while (looping)
+    {
+      for (select_record *s = di->start; (s = s->next); )
+	if (s->startup == start_thread_dsp)
+	  {
+	    if (peek_dsp (s, true))
+	      looping = false;
+	    if (di->stop_thread)
+	      {
+		select_printf ("stopping");
+		looping = false;
+		break;
+	      }
+	  }
+      if (!looping)
+	break;
+      cygwait (di->bye, sleep_time >> 3);
+      if (sleep_time < 80)
+	++sleep_time;
+      if (di->stop_thread)
+	break;
+    }
+  return 0;
+}
+
+static int
+start_thread_dsp (select_record *me, select_stuff *stuff)
+{
+  select_dsp_info *di = stuff->device_specific_dsp;
+  if (di->start)
+    me->h = *((select_dsp_info *) stuff->device_specific_dsp)->thread;
+  else
+    {
+      di->bye = me->fh->get_select_sem ();
+      if (di->bye)
+	DuplicateHandle (GetCurrentProcess (), di->bye,
+			 GetCurrentProcess (), &di->bye,
+			 0, 0, DUPLICATE_SAME_ACCESS);
+      else
+	di->bye = CreateSemaphore (&sec_none_nih, 0, INT32_MAX, NULL);
+      di->start = &stuff->start;
+      di->stop_thread = false;
+      di->thread = new cygthread (thread_dsp, di, "dspsel");
+      me->h = *di->thread;
+      if (!me->h)
+	return 0;
+    }
+  return 1;
+}
+
+static void
+dsp_cleanup (select_record *, select_stuff *stuff)
+{
+  select_dsp_info *di = (select_dsp_info *) stuff->device_specific_dsp;
+  if (!di)
+    return;
+  if (di->thread)
+    {
+      di->stop_thread = true;
+      ReleaseSemaphore (di->bye, get_obj_handle_count (di->bye), NULL);
+      di->thread->detach ();
+      CloseHandle (di->bye);
+    }
+  delete di;
+  stuff->device_specific_dsp = NULL;
+}
+
+select_record *
+fhandler_dev_dsp::select_read (select_stuff *stuff)
+{
+  if (!stuff->device_specific_dsp
+      && (stuff->device_specific_dsp = new select_dsp_info) == NULL)
+    return NULL;
+  select_record *s = stuff->start.next;
+  s->startup = start_thread_dsp;
+  s->peek = peek_dsp;
+  s->verify = verify_ok;
+  s->cleanup = dsp_cleanup;
+  s->read_selected = true;
+  s->read_ready = false;
+  return s;
+}
+
+select_record *
+fhandler_dev_dsp::select_write (select_stuff *stuff)
+{
+  if (!stuff->device_specific_dsp
+      && (stuff->device_specific_dsp = new select_dsp_info) == NULL)
+    return NULL;
+  select_record *s = stuff->start.next;
+  s->startup = start_thread_dsp;
+  s->peek = peek_dsp;
+  s->verify = verify_ok;
+  s->cleanup = dsp_cleanup;
+  s->write_selected = true;
+  s->write_ready = false;
+  return s;
+}
+
+select_record *
+fhandler_dev_dsp::select_except (select_stuff *stuff)
+{
+  if (!stuff->device_specific_dsp
+      && (stuff->device_specific_dsp = new select_dsp_info) == NULL)
+    return NULL;
+  select_record *s = stuff->start.next;
+  s->startup = start_thread_dsp;
+  s->peek = peek_dsp;
+  s->verify = verify_ok;
+  s->cleanup = dsp_cleanup;
+  s->except_selected = true;
+  s->except_ready = false;
+  return s;
+}
